import json
import json
import logging
import os
import tempfile
import webbrowser
from abc import ABC
from pathlib import Path
from typing import Optional, TYPE_CHECKING, Union, Any

import omf
import pandas as pd

if TYPE_CHECKING:
    from omf import Project
    from omfpandas.blockmodels.geometry import RegularGeometry, TensorGeometry

SUPPORTED_BM_TYPES = ['RegularBlockModel', 'TensorGridBlockModel']

PathLike = Union[str, Path, os.PathLike]


class OMFPandas(ABC):

    def __init__(self, filepath: PathLike):
        """Instantiate the OMFPandas object.

        Args:
            filepath (Path): Path to the OMF file.

        Raises:
            FileNotFoundError: If the OMF file does not exist.
            ValueError: If the file is not an OMF file.
        """
        self._logger = logging.getLogger(__class__.__name__)

        if not isinstance(filepath, Path):
            filepath = Path(filepath)

        if not filepath.suffix == '.omf':
            raise ValueError(f'File is not an OMF file: {filepath}')
        self.filepath: Path = filepath
        self.project: Optional[Project] = None
        if filepath.exists():
            self.project = omf.load(str(filepath))

    def __repr__(self):
        res: str = f"OMF file({self.filepath})"
        res += f"\nElements: {self.element_types}"
        return res

    def __str__(self):
        res: str = f"OMF file({self.filepath})"
        res += f"\nElements: {self.element_types}"
        return res

    @property
    def element_types(self) -> Optional[dict[str, Any]]:
        """Dictionary of elements keyed by name

        In the special case of a composite element, the key will be the composite name and
        the value will be a dictionary of child elements keyed by name.

        """
        _elements = self.project.elements if self.project else []
        if _elements:
            element_dict = {}
            for e in _elements:
                if hasattr(e, 'elements') and e.elements:
                    element_dict[e.name] = {child.name: child.__class__.__name__ for child in e.elements}
                else:
                    element_dict[e.name] = e.__class__.__name__
            return element_dict
        else:
            return {}

    @property
    def blockmodel_attributes(self) -> Optional[dict[str, Any]]:
        """Attributes for blockmodel elements, keyed by element name, including composites."""
        elements = {}
        for el in self.project.elements:
            if el.__class__.__name__ in ['TensorGridBlockModel', 'RegularBlockModel']:
                elements[el.name] = [a.name for a in el.attributes]
            elif hasattr(el, 'elements') and el.elements:
                composite_elements = {}
                for child in el.elements:
                    if child.__class__.__name__ in ['TensorGridBlockModel', 'RegularBlockModel']:
                        composite_elements[child.name] = [a.name for a in child.attributes]
                if composite_elements:
                    elements[el.name] = composite_elements

        return elements if elements else {}

    @property
    def changelog(self) -> Optional[pd.DataFrame]:
        """Return the change log as a DataFrame."""
        if 'changelog' not in self.project.metadata:
            return None
        return pd.DataFrame([json.loads(msg) for msg in self.project.metadata['changelog']])

    def get_element_by_name(self, element_name: str):
        """Get an element by its name.

        :param element_name: The name of the element to retrieve. Use dot notation for composite (e.g., Composite.BlockModel).
        :return:
        """
        composite_name = None
        if '.' in element_name:
            composite_name, element_name = element_name.split('.', 1)

        if composite_name:
            composite = [e for e in self.project.elements if e.name == composite_name]
            if not composite:
                raise ValueError(f"Composite '{composite_name}' not found in the OMF file: {self.filepath.name}.")
            composite = composite[0]
            element = [e for e in composite.elements if e.name == element_name]
            if not element:
                raise ValueError(f"Element '{element_name}' not found in composite '{composite_name}'.")
        else:
            element = [e for e in self.project.elements if e.name == element_name]
            if not element:
                raise ValueError(f"Element '{element_name}' not found in the OMF file: {self.filepath.name}.")
            elif len(element) > 1:
                raise ValueError(f"Multiple elements with the name '{element_name}' found in the OMF file: "
                                 f"{self.filepath.name}")

        return element[0]

    def get_element_attribute_names(self, element_name: str) -> list[str]:
        """Get the attribute names of an element.

        :param element_name: The name of the element to retrieve.  Use dot notation for elements in a composite.
        :return:
        """
        element = self.get_element_by_name(element_name)
        return [attr.name for attr in element.attributes]

    def get_bm_geometry(self, blockmodel_name: str) -> Union['RegularGeometry', 'TensorGeometry']:
        """Get the geometry of a BlockModel.

        Args:
            blockmodel_name (str): The name of the BlockModel to retrieve.

        Returns:
            TensorGeometry: The geometry of the BlockModel.
        """
        bm = self.get_element_by_name(blockmodel_name)
        if bm.__class__.__name__ == 'TensorGridBlockModel':
            from omfpandas.blockmodels.geometry import TensorGeometry
            return TensorGeometry.from_element(bm)
        elif bm.__class__.__name__ == 'RegularBlockModel':
            from omfpandas.blockmodels.geometry import RegularGeometry
            return RegularGeometry.from_element(bm)
        else:
            raise ValueError(
                f"Element '{blockmodel_name}' is not a supported BlockModel in the OMF file: {self.filepath}")

    def view_block_model_profile(self, blockmodel_name: str, query: Optional[str] = None):
        """View the profile of a BlockModel in the default web browser.

        Args:
            blockmodel_name (str): The name of the BlockModel to profile.
            query (str): A query defining the subset of the BlockModel.
        """

        el = self.get_element_by_name(blockmodel_name)
        filter_key: str = query if query else 'no_filter'

        if el.metadata.get('profile') is None:
            raise ValueError(f"BlockModel '{blockmodel_name}' has not been profiled.  "
                             f"Please run 'profile_blockmodel' first.")

        # JSON string containing the profile report
        profile_html: str = el.metadata['profile'][filter_key]

        # Create a temporary file
        with tempfile.NamedTemporaryFile(delete=False, suffix=".html") as temp_file:
            temp_file.write(profile_html.encode('utf-8'))
            temp_file_path = temp_file.name

        # Open the temporary file in the default web browser
        webbrowser.open(f"file://{temp_file_path}")



from typing import Union, TypeVar, Optional

import numpy as np
import pandas as pd
from omf import CategoryAttribute, NumericAttribute, CategoryColormap
from omf.blockmodel import BaseBlockModel, RegularBlockModel, TensorGridBlockModel
from pandas.core.dtypes.common import is_integer_dtype

from omfpandas.blockmodels.geometry import RegularGeometry, TensorGeometry
from omfpandas.utils.pandas_utils import is_nullable_integer_dtype, to_numpy_integer_dtype, to_nullable_integer_dtype, \
    parse_vars_from_expr

# generic type variable, used for type hinting, to indicate that the type is a subclass of BaseBlockModel
BM = TypeVar('BM', bound=BaseBlockModel)

SENTINEL_VALUE = -9  # TODO: possibly move to config file


def series_to_attribute(series: pd.Series) -> Union[CategoryAttribute, NumericAttribute]:

    # todo manage sorting - see attribute_to_series
    if isinstance(series.dtype, pd.CategoricalDtype):
        cat_map = {i: c for i, c in enumerate(series.cat.categories)}
        cat_col_map = CategoryColormap(indices=list(cat_map.keys()), values=list(cat_map.values()))
        attribute = CategoryAttribute(name=series.name, location="cells", array=np.array(series.cat.codes),
                                      categories=cat_col_map)
    else:
        # manage the sentinel / null placeholders
        # REF: https://github.com/gmggroup/omf-python/issues/59
        if is_nullable_integer_dtype(series):
            # set null_values and assign metadata
            data: pd.Series = series.fillna(SENTINEL_VALUE).pipe(to_numpy_integer_dtype)
            attribute = NumericAttribute(name=series.name, location="cells", array=data.values)
            attribute.metadata['null_value'] = SENTINEL_VALUE
        elif is_integer_dtype(series):
            attribute = NumericAttribute(name=series.name, location="cells", array=series.values)
            attribute.metadata['null_value'] = SENTINEL_VALUE
        else:
            attribute = NumericAttribute(name=series.name, location="cells", array=series.values)
            attribute.metadata['null_value'] = 'np.nan'
    return attribute


def attribute_to_series(attribute: Union[CategoryAttribute, NumericAttribute]) -> pd.Series:
    if isinstance(attribute, CategoryAttribute):
        return pd.Series(pd.Categorical.from_codes(codes=attribute.array.array.ravel(),
                                                   categories=attribute.categories.values,
                                                   ordered=False), name=attribute.name)
    else:
        # if an int with null_value in metadata then convert to a nullable int
        if attribute.metadata.get("null_value") and is_integer_dtype(attribute.array.array):
            return pd.Series(attribute.array.array.ravel(), name=attribute.name).pipe(
                to_nullable_integer_dtype).replace(
                SENTINEL_VALUE, pd.NA)
        return pd.Series(attribute.array.array.ravel(), name=attribute.name, dtype=attribute.array.array.dtype)


def get_attribute_by_name(blockmodel: BM, attr_name: str) -> Union[CategoryAttribute, NumericAttribute]:
    """Get the variable/attribute by its name from a BlockModel.

    Args:
        blockmodel (BlockModel): The BlockModel to get the data from.
        attr_name (str): The name of the attribute to retrieve.

    Returns:
        Union[CategoryAttribute, NumericAttribute]: The attribute with the given name.

    Raises:
        ValueError: If the variable is not found as cell data in the BlockModel or if multiple variables with the
        same name are found.
    """
    attrs = [sd for sd in blockmodel.attributes if sd.location == 'cells' and sd.name == attr_name]
    if not attrs:
        raise ValueError(f"Variable '{attr_name}' not found as cell data in the BlockModel: {blockmodel}")
    elif len(attrs) > 1:
        raise ValueError(f"Multiple variables with the name '{attr_name}' found in the BlockModel: {blockmodel}")
    return attrs[0]


def evaluate_calculated_attribute(blockmodel: BM, attr_name: str, calculated_expression: str,
                                  attributes_available: list[str]) -> pd.Series:
    """Evaluate a calculated attribute using the blockmodel and available attributes.

    Args:
        blockmodel (BlockModel): The BlockModel to read from.
        attr_name (str): The name of the calculated attribute.
        calculated_expression (str): The expression to evaluate.
        attributes_available (list[str]): List of available attributes in the BlockModel.

    Returns:
        pd.Series: The evaluated calculated attribute as a pandas Series.
    """
    local_dict = {attr: attribute_to_series(get_attribute_by_name(blockmodel, attr)) for attr in attributes_available}
    return pd.Series(eval(calculated_expression, {}, local_dict), name=attr_name)


def read_blockmodel_attributes(blockmodel: BM, attributes: Optional[list[str]] = None,
                               query: Optional[str] = None, index_filter: Optional[list[int]] = None) -> pd.DataFrame:
    """Read the attributes/variables from the BlockModel, including calculated attributes.

    Args:
        blockmodel (BlockModel): The BlockModel to read from.
        attributes (list[str]): The attributes to include in the DataFrame.
        query (str): The query to filter the DataFrame.
        index_filter (list[int]): List of integer indices to filter the DataFrame.

    Returns:
        pd.DataFrame: The DataFrame representing the attributes in the BlockModel.

    Raises:
        ValueError: If the attribute is not found in the BlockModel or if both query and index_filter are provided.
    """
    if query and index_filter:
        raise ValueError("Cannot use both query and index_filter at the same time.")

    # identify 'cell' variables in the file
    attributes_available = [v.name for v in blockmodel.attributes if v.location == 'cells']

    # Retrieve calculated attributes from metadata
    calculated_attributes: dict[str, str] = blockmodel.metadata.get('calculated_attributes', {})

    attributes: list[str] = attributes or (attributes_available + list(calculated_attributes.keys()))

    # check if the variables are available
    if not set(attributes).issubset(attributes_available + list(calculated_attributes.keys())):
        raise ValueError(
            f"Variables {set(attributes).difference(attributes_available + list(calculated_attributes.keys()))} "
            f"not found in the BlockModel.")

    int_index: np.ndarray = np.arange(blockmodel.num_cells)
    if query is not None:
        # parse out the attributes from the query using a package
        query_attrs = parse_vars_from_expr(query)
        # check if the attributes in the query are available
        if not set(query_attrs).issubset(attributes_available + list(calculated_attributes.keys())):
            raise ValueError(
                f"Variables {set(query_attrs).difference(attributes_available + list(calculated_attributes.keys()))} "
                f"not found in the BlockModel.")
        query_series: list = []
        for attr_name in query_attrs:
            if attr_name in calculated_attributes:
                query_series.append(
                    evaluate_calculated_attribute(blockmodel, attr_name, calculated_attributes[attr_name],
                                                  attributes_available))
            else:
                query_series.append(attribute_to_series(get_attribute_by_name(blockmodel, attr_name)))
        df_to_query: pd.DataFrame = pd.concat(query_series, axis=1)
        int_index = np.array(df_to_query.query(query).index)
    elif index_filter is not None:
        int_index = np.array(index_filter)

    # Loop over the variables
    chunks: list = []
    attr: str
    for attr in attributes:
        if attr in calculated_attributes:
            # Evaluate the calculated attribute
            calculated_series = evaluate_calculated_attribute(blockmodel, attr, calculated_attributes[attr],
                                                              attributes_available)
            chunks.append(calculated_series.iloc[int_index])
        else:
            attr: Union[CategoryAttribute, NumericAttribute] = get_attribute_by_name(blockmodel, attr)
            chunks.append(attribute_to_series(attr).iloc[int_index])

    # create the geometry index
    if isinstance(blockmodel, RegularBlockModel):
        geometry_index: pd.MultiIndex = RegularGeometry.from_element(blockmodel).to_multi_index()
    elif isinstance(blockmodel, TensorGridBlockModel):
        geometry_index: pd.MultiIndex = TensorGeometry.from_element(blockmodel).to_multi_index()
    else:
        raise ValueError(f"BlockModel type {blockmodel.__class__.__name__} not (yet) supported.")

    if (query is not None) or (index_filter is not None):
        # filter the index to match the int_index positional index
        geometry_index = geometry_index.take(int_index)

    res = pd.concat(chunks, axis=1)
    # res.index = geometry_index.to_frame().reset_index(drop=True).sort_values(by=['z', 'y', 'x']).set_index(geometry_index.names).index
    res.index = geometry_index
    return res if isinstance(res, pd.DataFrame) else res.to_frame()



import json
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from pathlib import Path
from typing import Union, Optional

import numpy as np
import pandas as pd

from omfpandas.blockmodels import multiindex_to_encoded_index

FloatArray = Union[np.ndarray, list[float], np.ndarray[float]]
Vector = Union[tuple[float, float, float], list[float, float, float]]
Point = Union[tuple[float, float, float], list[float, float, float]]
Triple = Union[tuple[float, float, float], list[float, float, float]]
MinMax = Union[tuple[float, float], list[float, float]]


# class CustomEncoder(json.JSONEncoder):
#     def default(self, obj):
#         if isinstance(obj, Vector3):
#             return obj.to_dict()
#         return super().default(obj)


@dataclass
class Geometry(ABC):
    """Base class for geometry objects.

    The geometry associated with omf block models are not defined by block centroids, and vary by block model type.
    In the pandas representation, the geometry is defined by the block centroids, so this class is used to
    define the geometry in terms of block centroids.
    Additionally, other properties of the geometry are defined here, such as the shape of the geometry.

    Attributes (in omf and pyvista) are stored in Fortran 'F' order, meaning that the last index changes the fastest.
    Hence the MultiIndex levels need to be sorted by 'z', 'y', 'x', to align with the Fortran order.
    This has x changing fastest, z changing slowest.

    """

    corner: Point
    axis_u: Vector
    axis_v: Vector
    axis_w: Vector
    _centroid_u: Optional[FloatArray] = field(default=None, init=False, repr=False)
    _centroid_v: Optional[FloatArray] = field(default=None, init=False, repr=False)
    _centroid_w: Optional[FloatArray] = field(default=None, init=False, repr=False)
    _shape: Optional[Point] = field(default=None, init=False, repr=False)
    _is_regular: Optional[bool] = field(default=None, init=False, repr=False)

    def to_summary_json(self) -> str:
        """Convert the geometry to a JSON string.

        Returns:
            str: The JSON string representing the geometry.
        """
        return json.dumps(self.summary)

    def to_json_file(self, json_filepath: Path) -> Path:
        """Write the Geometry to a JSON file.

        Args:
            json_filepath (Path): The path to write the JSON file.

        Returns:
            Path to the json file.
        """
        json_filepath.write_text(self.to_json())
        return json_filepath

    @abstractmethod
    def __repr__(self):
        pass

    @abstractmethod
    def __str__(self):
        pass

    @property
    @abstractmethod
    def is_regular(self) -> bool:
        pass

    @property
    @abstractmethod
    def centroid_u(self) -> np.ndarray[float]:
        pass

    @property
    @abstractmethod
    def centroid_v(self) -> np.ndarray[float]:
        pass

    @property
    @abstractmethod
    def centroid_w(self) -> np.ndarray[float]:
        pass

    @property
    def num_cells(self) -> int:
        return int(np.prod(self.shape))

    @property
    def shape(self) -> Triple:
        if self._shape is None:
            self._shape = (
                len(self.centroid_u),
                len(self.centroid_v),
                len(self.centroid_w),
            )
        return self._shape

    @property
    @abstractmethod
    def extents(self) -> tuple[MinMax, MinMax, MinMax]:
        pass

    @property
    def bounding_box(self) -> tuple[MinMax, MinMax]:
        return self.extents[0], self.extents[1]

    @property
    @abstractmethod
    def summary(self) -> dict:
        pass

    @classmethod
    @abstractmethod
    def from_element(cls, element) -> "Geometry":
        pass

    @classmethod
    @abstractmethod
    def from_multi_index(cls, index: pd.MultiIndex):
        pass

    @abstractmethod
    def to_multi_index(self) -> pd.MultiIndex:
        pass

    @abstractmethod
    def nearest_centroid_lookup(self, x: float, y: float, z: float) -> Point:
        pass


@dataclass
class RegularGeometry(Geometry):
    """Regular geometry data class.

    Regular Geometry applies to an omf.v1 VolumeElement or an omf.v2 RegularBlockModel.
    """

    block_size: Triple
    _shape: Triple = field(default=None, init=False, repr=False)

    def __init__(
        self,
        corner: Point,
        axis_u: Vector,
        axis_v: Vector,
        axis_w: Vector,
        block_size: Triple,
        shape: Triple = None,
    ):
        self.corner = corner
        self.axis_u = axis_u
        self.axis_v = axis_v
        self.axis_w = axis_w
        self.block_size = block_size
        self._shape = shape

    def __repr__(self):
        return f"RegularGeometry: {self.summary}"

    def __str__(self):
        return f"RegularGeometry: {self.summary}"

    @property
    def is_regular(self) -> bool:
        return True

    @property
    def centroid_u(self) -> np.ndarray[float]:
        if self._centroid_u is None:
            self._centroid_u = np.arange(
                self.corner[0] + self.block_size[0] / 2,
                self.corner[0] + self.block_size[0] * self.shape[0],
                self.block_size[0],
            )
        return self._centroid_u

    @property
    def centroid_v(self) -> np.ndarray[float]:
        if self._centroid_v is None:
            self._centroid_v = np.arange(
                self.corner[1] + self.block_size[1] / 2,
                self.corner[1] + self.block_size[1] * self.shape[1],
                self.block_size[1],
            )
        return self._centroid_v

    @property
    def centroid_w(self) -> np.ndarray[float]:
        if self._centroid_w is None:
            self._centroid_w = np.arange(
                self.corner[2] + self.block_size[2] / 2,
                self.corner[2] + self.block_size[2] * self.shape[2],
                self.block_size[2],
            )
        return self._centroid_w

    @property
    def shape(self) -> Triple:
        return self._shape

    @shape.setter
    def shape(self, value: Triple):
        self._shape = value

    @property
    def extents(self) -> tuple[MinMax, MinMax, MinMax]:
        return (
            (
                float(self.centroid_u[0] - self.block_size[0] / 2),
                float(self.centroid_u[-1] + self.block_size[0] / 2),
            ),
            (
                float(self.centroid_v[0] - self.block_size[1] / 2),
                float(self.centroid_v[-1] + self.block_size[1] / 2),
            ),
            (
                float(self.centroid_w[0] - self.block_size[2] / 2),
                float(self.centroid_w[-1] + self.block_size[2] / 2),
            ),
        )

    @property
    def summary(self) -> dict:
        return {
            "corner": tuple(self.corner),
            "axis_u": tuple(self.axis_u),
            "axis_v": tuple(self.axis_v),
            "axis_w": tuple(self.axis_w),
            "block_size": self.block_size,
            "shape": self.shape,
            "is_regular": self.is_regular,
            "extents": self.extents,
            "bounding_box": self.bounding_box,
        }

    @classmethod
    def from_element(cls, element) -> "RegularGeometry":
        from omf import RegularBlockModel

        if not isinstance(element, RegularBlockModel):
            raise ValueError("Element must be an instance of omf.RegularBlockModel")
        return cls(
            element.corner,
            element.axis_u,
            element.axis_v,
            element.axis_w,
            element.block_size,
            shape=element.block_count,
        )

    @classmethod
    def from_multi_index(
        cls,
        index: pd.MultiIndex,
        axis_u: Vector = (1, 0, 0),
        axis_v: Vector = (0, 1, 0),
        axis_w: Vector = (0, 0, 1),
    ) -> "RegularGeometry":

        # check that the index contains the expected levels
        if not {"x", "y", "z"}.issubset(index.names):
            raise ValueError("Index must contain the levels 'x', 'y', 'z'.")

        x = index.get_level_values("x").unique()
        y = index.get_level_values("y").unique()
        z = index.get_level_values("z").unique()

        # check the block sizes are unique
        dx = np.unique(np.diff(x))
        dy = np.unique(np.diff(y))
        dz = np.unique(np.diff(z))

        # Use the minimum difference as the block size
        block_size = float(dx.min()), float(dy.min()), float(dz.min())

        # Calculate the shape based on the full range of coordinates
        full_shape = (
            len(np.arange(x.min(), x.max() + block_size[0], block_size[0])),
            len(np.arange(y.min(), y.max() + block_size[1], block_size[1])),
            len(np.arange(z.min(), z.max() + block_size[2], block_size[2])),
        )

        corner_x = x.min() - block_size[0] / 2
        corner_y = y.min() - block_size[1] / 2
        corner_z = z.min() - block_size[2] / 2

        # Create the volume
        return cls(
            corner=(corner_x, corner_y, corner_z),
            axis_u=axis_u,
            axis_v=axis_v,
            axis_w=axis_w,
            block_size=block_size,
            shape=full_shape,
        )

    @classmethod
    def from_extents(
        cls,
        extents: tuple[MinMax, MinMax, MinMax],
        block_size: Triple,
        axis_u: Vector = (1, 0, 0),
        axis_v: Vector = (0, 1, 0),
        axis_w: Vector = (0, 0, 1),
    ) -> "RegularGeometry":
        """Create a RegularGeometry from extents."""
        min_x, max_x = extents[0]
        min_y, max_y = extents[1]
        min_z, max_z = extents[2]

        corner = (
            min_x - block_size[0] / 2,
            min_y - block_size[1] / 2,
            min_z - block_size[2] / 2,
        )
        shape = (
            int((max_x - min_x) / block_size[0]),
            int((max_y - min_y) / block_size[1]),
            int((max_z - min_z) / block_size[2]),
        )

        return cls(corner, axis_u, axis_v, axis_w, block_size, shape)

    def to_json(self) -> str:
        """Convert the full geometry to a JSON string."""
        data = {
            "corner": list(self.corner),
            "axis_u": list(self.axis_u),
            "axis_v": list(self.axis_v),
            "axis_w": list(self.axis_w),
            "block_size": list(self.block_size),
            "shape": list(self.shape),
        }
        return json.dumps(data)

    @classmethod
    def from_json(cls, json_str: str) -> "RegularGeometry":
        """Deserialize a JSON string to a full geometry object."""
        data = json.loads(json_str)
        return cls(
            corner=list(data["corner"]),
            axis_u=list(data["axis_u"]),
            axis_v=list(data["axis_v"]),
            axis_w=list(data["axis_w"]),
            block_size=list(data["block_size"]),
            shape=list(data["shape"]),
        )

    def to_multi_index(self) -> pd.MultiIndex:
        """Convert a RegularGeometry to a MultiIndex.

        The MultiIndex will have the following levels:
        - x: The x coordinates of the cell centres
        - y: The y coordinates of the cell centres
        - z: The z coordinates of the cell centres
        """

        """Returns a pd.MultiIndex for the regular blockmodel element, accounting for rotation.

        Args:
            blockmodel (BaseBlockModel): The regular BlockModel to get the index from.

        Returns:
            pd.MultiIndex: The MultiIndex representing the blockmodel element geometry.
        """
        ox, oy, oz = self.corner
        dx, dy, dz = self.block_size
        nx, ny, nz = self.shape

        # Calculate the coordinates of the block centers
        x = ox + (np.arange(nx) + 0.5) * dx
        y = oy + (np.arange(ny) + 0.5) * dy
        z = oz + (np.arange(nz) + 0.5) * dz

        # Create a grid of coordinates
        xx, yy, zz = np.meshgrid(x, y, z, indexing="ij")

        # Flatten the grid
        centroids = np.vstack([xx.ravel(), yy.ravel(), zz.ravel()])

        # Rotation axes
        u, v, w = self.axis_u, self.axis_v, self.axis_w

        # Create rotation matrix
        rotation_matrix = np.array([u, v, w]).T

        # Apply rotation
        rotated_centroids = rotation_matrix @ centroids

        # Create a MultiIndex
        index = pd.MultiIndex.from_arrays(
            [rotated_centroids[0], rotated_centroids[1], rotated_centroids[2]],
            names=["x", "y", "z"],
        )

        # Sort the MultiIndex by x, y, z levels
        return index.sortlevel(level=["x", "y", "z"])[0]

    def to_encoded_index(self) -> pd.Index:
        """Convert a RegularGeometry to an encoded integer index

        The integer index is encoded to preserve the spatial position.

        Use the coordinate_hashing.hashed_index_to_multiindex function to convert it back to x, y, z pd.MultiIndex

        Returns:

        """
        return multiindex_to_encoded_index(self.to_multi_index())

    def nearest_centroid_lookup(self, x: float, y: float, z: float) -> Point:
        """Find the nearest centroid for provided x, y, z points.

        Args:
            x (float): X coordinate.
            y (float): Y coordinate.
            z (float): Z coordinate.

        Returns:
            Point3: The coordinates of the nearest centroid.
        """

        reference_centroid: Point = (
            self.centroid_u[0],
            self.centroid_v[0],
            self.centroid_w[0],
        )
        dx, dy, dz = self.block_size
        ref_x, ref_y, ref_z = reference_centroid

        nearest_x = round((x - ref_x) / dx) * dx + ref_x
        nearest_y = round((y - ref_y) / dy) * dy + ref_y
        nearest_z = round((z - ref_z) / dz) * dz + ref_z

        return nearest_x, nearest_y, nearest_z


@dataclass
class TensorGeometry(Geometry):
    """Tensor geometry data class.

    Applicable for on omf.v2 TensorGridBlockModel.
    """

    tensor_u: FloatArray
    tensor_v: FloatArray
    tensor_w: FloatArray
    _block_sizes: Optional[Point] = field(default=None, init=False)

    def __repr__(self):
        return f"TensorGeometry: {self.summary}"

    def __str__(self):
        return f"TensorGeometry: {self.summary}"

    @property
    def is_regular(self) -> bool:
        return (
            np.allclose(self.tensor_u, self.tensor_u[0])
            and np.allclose(self.tensor_v, self.tensor_v[0])
            and np.allclose(self.tensor_w, self.tensor_w[0])
        )

    @property
    def centroid_u(self) -> np.ndarray[float]:
        if self._centroid_u is None:
            self._centroid_u = (
                self.corner[0] + np.cumsum(self.tensor_u) - self.tensor_u / 2
            )
        return self._centroid_u

    @property
    def centroid_v(self) -> np.ndarray[float]:
        if self._centroid_v is None:
            self._centroid_v = (
                self.corner[0] + np.cumsum(self.tensor_v) - self.tensor_v / 2
            )
        return self._centroid_v

    @property
    def centroid_w(self) -> np.ndarray[float]:
        if self._centroid_w is None:
            self._centroid_w = (
                self.corner[0] + np.cumsum(self.tensor_w) - self.tensor_w / 2
            )
        return self._centroid_w

    @property
    def block_sizes(self) -> Triple:
        if self._block_sizes is None:
            # Create a grid of all possible combinations
            grid = np.array(
                np.meshgrid(self.tensor_u, self.tensor_v, self.tensor_w)
            ).T.reshape(-1, 3)
            # Find unique combinations
            unique_block_sizes = np.unique(grid, axis=0)
            self._block_sizes = [tuple(map(float, size)) for size in unique_block_sizes]
        return self._block_sizes

    @property
    def extents(self) -> tuple[MinMax, MinMax, MinMax]:
        return (
            (
                float(self.centroid_u[0] - self.tensor_u[0] / 2),
                float(self.centroid_u[-1] + self.tensor_u[0] / 2),
            ),
            (
                float(self.centroid_v[0] - self.tensor_v[0] / 2),
                float(self.centroid_v[-1] + self.tensor_v[0] / 2),
            ),
            (
                float(self.centroid_w[0] - self.tensor_w[0] / 2),
                float(self.centroid_w[-1] + self.tensor_w[0] / 2),
            ),
        )

    @property
    def summary(self) -> dict:
        return {
            "corner": tuple(self.corner),
            "axis_u": tuple(self.axis_u),
            "axis_v": tuple(self.axis_v),
            "axis_w": tuple(self.axis_w),
            "block_sizes": self.block_sizes,
            "shape": self.shape,
            "is_regular": self.is_regular,
            "extents": self.extents,
            "bounding_box": self.bounding_box,
        }

    @classmethod
    def from_element(cls, element) -> "TensorGeometry":
        from omf import TensorGridBlockModel

        if not isinstance(element, TensorGridBlockModel):
            raise ValueError("Element must be an instance of omf.TensorGridBlockModel")

        return cls(
            element.corner,
            element.axis_u,
            element.axis_v,
            element.axis_w,
            element.tensor_u,
            element.tensor_v,
            element.tensor_w,
        )

    @classmethod
    def from_multi_index(cls, index: pd.MultiIndex) -> "TensorGeometry":
        # check that the index contains the expected levels
        level_names: list[str] = ["x", "y", "z", "dx", "dy", "dz"]
        if not set(level_names).issubset(index.names):
            raise ValueError(f"Index must contain the levels {level_names}.")

        x = index.get_level_values("x").unique()
        y = index.get_level_values("y").unique()
        z = index.get_level_values("z").unique()

        # Get the shape of the original 3D arrays
        shape = (len(x), len(y), len(z))

        # Reshape the ravelled index back into the original shapes
        tensor_u = index.get_level_values("dx").values.reshape(shape, order="F")[
            :, 0, 0
        ]
        tensor_v = index.get_level_values("dy").values.reshape(shape, order="F")[
            0, :, 0
        ]
        tensor_w = index.get_level_values("dz").values.reshape(shape, order="F")[
            0, 0, :
        ]

        origin_x = x.min() - tensor_u[0] / 2
        origin_y = y.min() - tensor_v[0] / 2
        origin_z = z.min() - tensor_w[0] / 2

        # Create the volume
        return cls(
            corner=(origin_x, origin_y, origin_z),
            axis_u=(1, 0, 0),
            axis_v=(0, 1, 0),
            axis_w=(0, 0, 1),
            tensor_u=tensor_u,
            tensor_v=tensor_v,
            tensor_w=tensor_w,
        )

    def to_json(self) -> str:
        """Convert the full geometry to a JSON string."""
        data = {
            "corner": list(self.corner),
            "axis_u": list(self.axis_u),
            "axis_v": list(self.axis_v),
            "axis_w": list(self.axis_w),
            "tensor_u": self.tensor_u.tolist(),
            "tensor_v": self.tensor_v.tolist(),
            "tensor_w": self.tensor_w.tolist(),
        }
        return json.dumps(data)

    @classmethod
    def from_json(cls, json_str: str) -> "TensorGeometry":
        """Deserialize a JSON string to a full geometry object."""
        data = json.loads(json_str)
        return cls(
            corner=list(data["corner"]),
            axis_u=list(data["axis_u"]),
            axis_v=list(data["axis_v"]),
            axis_w=list(data["axis_w"]),
            tensor_u=np.array(data["tensor_u"]),
            tensor_v=np.array(data["tensor_v"]),
            tensor_w=np.array(data["tensor_w"]),
        )

    def to_multi_index(self) -> pd.MultiIndex:
        """Convert a TensorGeometry to a MultiIndex.

        The MultiIndex will have the following levels:
        - x: The x coordinates of the cell centres
        - y: The y coordinates of the cell centres
        - z: The z coordinates of the cell centres
        - dx: The x cell sizes
        - dy: The y cell sizes
        - dz: The z cell sizes
        """

        """Returns a pd.MultiIndex for the tensor blockmodel element.

        Args:
            blockmodel (BlockModel): The tensor BlockModel to get the index from.

        Returns:
            pd.MultiIndex: The MultiIndex representing the blockmodel element geometry.
        """
        ox, oy, oz = self.corner

        # Make coordinates (points) along each axis, i, j, k
        i = ox + np.cumsum(self.tensor_u)
        i = np.insert(i, 0, ox)
        j = oy + np.cumsum(self.tensor_v)
        j = np.insert(j, 0, oy)
        k = oz + np.cumsum(self.tensor_w)
        k = np.insert(k, 0, oz)

        # convert to centroids
        x, y, z = (i[1:] + i[:-1]) / 2, (j[1:] + j[:-1]) / 2, (k[1:] + k[:-1]) / 2
        xx, yy, zz = np.meshgrid(x, y, z, indexing="ij")

        # Calculate dx, dy, dz
        dxx, dyy, dzz = np.meshgrid(
            self.tensor_u, self.tensor_v, self.tensor_w, indexing="ij"
        )

        # TODO: consider rotation

        index: pd.MultiIndex = pd.MultiIndex.from_arrays(
            [xx.ravel(), yy.ravel(), zz.ravel(), dxx.ravel(), dyy.ravel(), dzz.ravel()],
            names=["x", "y", "z", "dx", "dy", "dz"],
        )

        # Sort the MultiIndex by x, y, z levels
        return index.sortlevel(level=["x", "y", "z"])[0]

    def nearest_centroid_lookup(self, x: float, y: float, z: float) -> Point:
        """Find the nearest centroid for provided x, y, z points.

        Args:
            x (float): X coordinate.
            y (float): Y coordinate.
            z (float): Z coordinate.

        Returns:
            Point3: The coordinates of the nearest centroid.
        """

        # This only works for regular geometries in a tensor format - needs to be extended.
        if not self.is_regular:
            raise ValueError(
                "TensorGeometry is not regular. Cannot perform nearest centroid lookup."
            )

        reference_centroid: Point = (
            self.centroid_u[0],
            self.centroid_v[0],
            self.centroid_w[0],
        )
        dx, dy, dz = self.block_sizes[0]
        ref_x, ref_y, ref_z = reference_centroid

        nearest_x = round((x - ref_x) / dx) * dx + ref_x
        nearest_y = round((y - ref_y) / dy) * dy + ref_y
        nearest_z = round((z - ref_z) / dz) * dz + ref_z

        return nearest_x, nearest_y, nearest_z



import tokenize
from io import StringIO
from token import STRING
from typing import Literal

import pandas as pd
import numpy as np
from pandas import CategoricalDtype


def is_nullable_integer_dtype(series: pd.Series) -> bool:
    """

    Args:
        series: The series

    Returns:
        bool: True if series contains nullable integer
    """

    return True if str(series.dtype)[0] == "I" else False


def to_nullable_integer_dtype(series: pd.Series) -> pd.Series:
    """ Convert an int series to a nullable integer dtype

    Args:
        series: The series

    Returns:
        pd.Series: The series with nullable dtype
    """

    return series.astype(str(series.dtype).replace("i", "I")) if is_nullable_integer_dtype(series) else series


def to_numpy_integer_dtype(series: pd.Series) -> pd.Series:
    """ Convert a nullable int series to a numpy integer dtype

    Args:
        series: The series

    Returns:
        pd.Series: The series with nullable dtype
    """

    return series.astype(str(series.dtype).replace("I", "i")) if is_nullable_integer_dtype(series) else series


def parse_vars_from_expr(expr: str) -> list[str]:
    """ Parse variables from a pandas query expression string.

    Args:
        expr: The expression string

    Returns:
        list[str]: The list of variables
    """
    variables = set()
    tokens = tokenize.generate_tokens(StringIO(expr).readline)
    logical_operators = {'and', 'or', '&', '|'}
    inside_backticks = False
    current_var = []

    for token in tokens:
        if token.string == '`':
            if inside_backticks:
                # End of backtick-enclosed variable
                variables.add(' '.join(current_var))
                current_var = []
            inside_backticks = not inside_backticks
        elif inside_backticks:
            if token.type in {tokenize.NAME, STRING}:
                current_var.append(token.string)
        elif token.type == tokenize.NAME and token.string not in logical_operators:
            variables.add(token.string)

    return list(variables)


def create_test_blockmodel(shape: tuple[int, int, int],
                           block_size: tuple[float, float, float],
                           corner: tuple[float, float, float],
                           is_tensor=False) -> pd.DataFrame:
    """
    Create a test blockmodel DataFrame.

    Args:
        shape: Shape of the block model (x, y, z).
        block_size: Size of each block (x, y, z).
        corner: The lower left (minimum) corner of the block model.
        is_tensor: If True, create a tensor block model. Default is False, which creates a regular block model.
            The MultiIndex levels for a regular model are x, y, z.  For a tensor model they are x, y, z, dx, dy, dz.
            The tensor model created is a special case where dx, dy, dz are the same for all blocks.

    Returns:
    pd.DataFrame: DataFrame containing the block model data.

    """

    num_blocks = np.prod(shape)

    # Generate the coordinates for the block model
    x_coords = np.arange(corner[0] + block_size[0] / 2, corner[0] + shape[0] * block_size[0], block_size[0])
    y_coords = np.arange(corner[1] + block_size[1] / 2, corner[1] + shape[1] * block_size[1], block_size[1])
    z_coords = np.arange(corner[2] + block_size[2] / 2, corner[2] + shape[2] * block_size[2], block_size[2])

    # Create a meshgrid of coordinates
    xx, yy, zz = np.meshgrid(x_coords, y_coords, z_coords, indexing='ij')

    # Flatten the coordinates
    xx_flat_c = xx.ravel(order='C')
    yy_flat_c = yy.ravel(order='C')
    zz_flat_c = zz.ravel(order='C')

    # Create the attributes
    c_order_xyz = np.arange(num_blocks)

    # assume the surface of the highest block is the topo surface
    surface_rl = corner[2] + shape[2] * block_size[2]

    # Create the DataFrame
    df = pd.DataFrame({
        'x': xx_flat_c,
        'y': yy_flat_c,
        'z': zz_flat_c,
        'c_style_xyz': c_order_xyz})

    # Set the index to x, y, z
    df.set_index(keys=['x', 'y', 'z'], inplace=True)
    df.sort_index(level=['x', 'y', 'z'], inplace=True)
    df.sort_index(level=['z', 'y', 'x'], inplace=True)
    df['f_style_zyx'] = c_order_xyz
    df.sort_index(level=['x', 'y', 'z'], inplace=True)

    df['depth'] = surface_rl - zz_flat_c

    # Check the ordering - confirm that the c_order_xyz and f_order_zyx columns are in the correct order
    assert np.array_equal(df.sort_index(level=['x', 'y', 'z'])['c_style_xyz'].values, np.arange(num_blocks))
    assert np.array_equal(df.sort_index(level=['z', 'y', 'x'])['f_style_zyx'].values, np.arange(num_blocks))

    # Check the depth using a pandas groupby
    depth_group = df.groupby('z')['depth'].unique().apply(lambda x: x[0]).sort_index(ascending=False)
    assert np.all(surface_rl - depth_group.diff().index == depth_group.values)

    if is_tensor:
        # Create the dx, dy, dz levels
        df['dx'] = block_size[0]
        df['dy'] = block_size[1]
        df['dz'] = block_size[2]

        # Set the index to x, y, z, dx, dy, dz
        df.set_index(keys=['dx', 'dy', 'dz'], append=True, inplace=True)

    return df


def aggregate(df: pd.DataFrame, agg_dict: dict, cat_treatment: Literal['majority', 'proportions'] = 'majority',
              proportions_as_columns: bool = False) -> pd.DataFrame:
    """
    Aggregate a DataFrame using a provided dictionary.

    Args:
        df: The DataFrame to aggregate.
        agg_dict: A dictionary where keys are the columns to be aggregated and values are the weight columns.
        cat_treatment: A string indicating how to treat categorical columns.
                       'majority' returns the majority category, 'proportions' returns the proportions of each category.
        proportions_as_columns: A boolean indicating whether to return category proportions as separate columns.

    Returns:
        pd.DataFrame: The aggregated DataFrame with columns in the same order as the incoming DataFrame.
    """
    result = {}
    weight_columns = set(agg_dict.values())

    for weight_col in weight_columns:
        # Get columns that share the same weight column
        cols_with_weight = [col for col, w_col in agg_dict.items() if w_col == weight_col]
        if cols_with_weight:
            weights = df[weight_col].values
            weighted_values = df[cols_with_weight].values * weights[:, np.newaxis]
            aggregated_values = np.sum(weighted_values, axis=0) / np.sum(weights)
            result.update({col: aggregated_values[i] for i, col in enumerate(cols_with_weight)})

    # Sum columns that are not in the agg_dict
    for col in df.columns:
        if col not in agg_dict:
            if isinstance(df[col].dtype, CategoricalDtype):
                if cat_treatment == 'majority':
                    result[col] = df[col].mode()[0]  # Get the majority category
                elif cat_treatment == 'proportions':
                    proportions = df[col].value_counts(normalize=True).to_dict()
                    if proportions_as_columns:
                        for cat, prop in proportions.items():
                            result[f"{col}_{cat}"] = prop
                    else:
                        result[col] = proportions
            else:
                result[col] = df[col].sum()

    # Create a DataFrame from the result dictionary
    aggregated_df = pd.DataFrame([result])

    # Manage the final column order
    if proportions_as_columns:
        # loop through the columns and add them, extending with cat classes
        final_columns = []
        for col in df.columns:
            if col in result:
                final_columns.append(col)
            elif isinstance(df[col].dtype, CategoricalDtype) and cat_treatment == 'proportions':
                for cat in df[col].cat.categories:
                    final_columns.append(f"{col}_{cat}")
        aggregated_df = aggregated_df[final_columns]
    else:
        # Ensure the columns are in the same order as the incoming DataFrame
        aggregated_df = aggregated_df[df.columns]

    return aggregated_df


"""
OMF uses C-order, sorting by x, y, z, while PyVista uses Fortran-order, sorting by z, y, x.
Our Pandas MultiIndex is sorted by x, y, z to align with OMF.
"""


from typing import Optional, Union

import numpy as np
import pandas as pd
from omf import TensorGridBlockModel, RegularBlockModel, NumericAttribute, CategoryAttribute

from omfpandas.blockmodels.attributes import read_blockmodel_attributes, BM, series_to_attribute
from omfpandas.blockmodels.geometry import RegularGeometry, TensorGeometry

import pyvista as pv

def df_to_blockmodel(df: pd.DataFrame, blockmodel_name: str) -> Union[RegularBlockModel, TensorGridBlockModel]:
    """
    Get the appropriate function to convert a DataFrame to a BlockModel.

    Args:
        df (pd.DataFrame): The DataFrame to convert to a RegularBlockModel.
        blockmodel_name (str): The name of the RegularBlockModel.

    Returns:
        The RegularBlockModel|TensorGridBlockModel representing the DataFrame.
    """

    if 'x' not in df.index.names and 'y' not in df.index.names and 'z' not in df.index.names:
        raise ValueError("Dataframe must have centroid coordinates (x, y, z) in the index.")
    elif 'dx' in df.index.names and 'dy' in df.index.names and 'dz' in df.index.names:
        return df_to_tensor_bm(df=df, blockmodel_name=blockmodel_name)
    else:
        return df_to_regular_bm(df=df, blockmodel_name=blockmodel_name)


def blockmodel_to_df(blockmodel: BM,
                     variables: Optional[list[str]] = None,
                     query: Optional[str] = None,
                     index_filter: Optional[list[int]] = None) -> pd.DataFrame:
    """Convert regular block model to a DataFrame.

    Args:
        blockmodel (BlockModel): The BlockModel to convert.
        variables (Optional[list[str]]): The variables to include in the DataFrame. If None, all variables are included.
        query (Optional[str]): The query to filter the DataFrame.
        index_filter (Optional[list[int]]): List of integer indices to filter the DataFrame.

    Returns:
        pd.DataFrame: The DataFrame representing the BlockModel.
    """
    # read the data
    df: pd.DataFrame = read_blockmodel_attributes(blockmodel, attributes=variables, query=query,
                                                  index_filter=index_filter)
    return df


def df_to_regular_bm(df: pd.DataFrame, blockmodel_name: str) -> RegularBlockModel:
    """Convert a DataFrame to a RegularBlockModel.

    Args:
        df (pd.DataFrame): The DataFrame to convert to a RegularBlockModel.
        blockmodel_name (str): The name of the RegularBlockModel.

    Returns:
        RegularBlockModel: The RegularBlockModel representing the DataFrame.
    """

    # Sort the dataframe to align with the omf spec - 'C' order
    df.sort_index(level=['x', 'y', 'z'])

    # Create the block model and geometry
    blockmodel = RegularBlockModel(name=blockmodel_name)
    geometry: RegularGeometry = RegularGeometry.from_multi_index(df.index)
    blockmodel.corner = geometry.corner
    blockmodel.axis_u = geometry.axis_u
    blockmodel.axis_v = geometry.axis_v
    blockmodel.axis_w = geometry.axis_w
    blockmodel.block_count = list(geometry.shape)
    blockmodel.block_size = list(geometry.block_size)
    blockmodel.cbc = [1] * geometry.num_cells

    # add the data
    attrs: list[Union[NumericAttribute, CategoryAttribute]] = []
    for variable in df.columns:
        attribute = series_to_attribute(df[variable])

        attrs.append(attribute)
    blockmodel.attributes = attrs
    blockmodel.validate()

    return blockmodel


def df_to_tensor_bm(df: pd.DataFrame, blockmodel_name: str) -> BM:
    """Write a DataFrame to a BlockModel.

    Args:
        df (pd.DataFrame): The DataFrame to convert to a BlockModel.
        blockmodel_name (str): The name of the BlockModel.
        created.

    Returns:
        BlockModel: The BlockModel representing the DataFrame.
    """

    # Sort the dataframe to align with the omf spec - 'C' order
    df.sort_index(level=['x', 'y', 'z'], inplace=True)

    # Create the blockmodel and geometry

    # if is_tensor:
    geometry: TensorGeometry = TensorGeometry.from_multi_index(df.index)

    blockmodel: BM = TensorGridBlockModel(name=blockmodel_name)
    # assign the geometry properties
    blockmodel.corner = geometry.corner
    blockmodel.axis_u = geometry.axis_u
    blockmodel.axis_v = geometry.axis_v
    blockmodel.axis_w = geometry.axis_w
    blockmodel.tensor_u = geometry.tensor_u
    blockmodel.tensor_v = geometry.tensor_v
    blockmodel.tensor_w = geometry.tensor_w

    # add the data
    attrs: list[Union[NumericAttribute, CategoryAttribute]] = []
    for variable in df.columns:
        attribute = series_to_attribute(df[variable])

        attrs.append(attribute)
    blockmodel.attributes = attrs
    blockmodel.validate()

    return blockmodel


def df_to_pv_structured_grid(df: pd.DataFrame) -> 'pv.StructuredGrid':
    import pyvista as pv

    # ensure the dataframe is sorted by z, y, x, since Pyvista uses 'F' order.
    df = df.sort_index(level=['z', 'y', 'x'])

    # Get the unique x, y, z coordinates (centroids)
    x_centroids = df.index.get_level_values('x').unique()
    y_centroids = df.index.get_level_values('y').unique()
    z_centroids = df.index.get_level_values('z').unique()

    # Calculate the cell size (assuming all cells are of equal size)
    dx = np.diff(x_centroids)[0]
    dy = np.diff(y_centroids)[0]
    dz = np.diff(z_centroids)[0]

    # Calculate the grid points
    x_points = np.concatenate([x_centroids - dx / 2, x_centroids[-1:] + dx / 2])
    y_points = np.concatenate([y_centroids - dy / 2, y_centroids[-1:] + dy / 2])
    z_points = np.concatenate([z_centroids - dz / 2, z_centroids[-1:] + dz / 2])

    # Create the 3D grid of points
    x, y, z = np.meshgrid(x_points, y_points, z_points, indexing='ij')

    # Create a StructuredGrid object
    grid = pv.StructuredGrid(x, y, z)

    # Add the data from the DataFrame to the grid
    for column in df.columns:
        grid.cell_data[column] = df[column].values

    return grid


def df_to_pv_unstructured_grid(df: pd.DataFrame) -> 'pv.UnstructuredGrid':
    """
    Requires the index to be a pd.MultiIndex with names ['x', 'y', 'z', 'dx', 'dy', 'dz'].
    :return:
    """

    # ensure the dataframe is sorted by z, y, x, since Pyvista uses 'F' order.
    blocks = df.reset_index().sort_values(['z', 'y', 'x'])

    # Get the x, y, z coordinates and cell dimensions
    # if no dims are passed, estimate them
    if 'dx' not in blocks.columns:
        dx, dy, dz = df.common_block_size()
        blocks['dx'] = dx
        blocks['dy'] = dy
        blocks['dz'] = dz

    x, y, z, dx, dy, dz = (blocks[col].values for col in blocks.columns if col in ['x', 'y', 'z', 'dx', 'dy', 'dz'])
    blocks.set_index(['x', 'y', 'z', 'dx', 'dy', 'dz'], inplace=True)
    # Create the cell points/vertices
    # REF: https://github.com/OpenGeoVis/PVGeo/blob/main/PVGeo/filters/voxelize.py

    n_cells = len(x)

    # Generate cell nodes for all points in data set
    # - Bottom
    c_n1 = np.stack(((x - dx / 2), (y - dy / 2), (z - dz / 2)), axis=1)
    c_n2 = np.stack(((x + dx / 2), (y - dy / 2), (z - dz / 2)), axis=1)
    c_n3 = np.stack(((x - dx / 2), (y + dy / 2), (z - dz / 2)), axis=1)
    c_n4 = np.stack(((x + dx / 2), (y + dy / 2), (z - dz / 2)), axis=1)
    # - Top
    c_n5 = np.stack(((x - dx / 2), (y - dy / 2), (z + dz / 2)), axis=1)
    c_n6 = np.stack(((x + dx / 2), (y - dy / 2), (z + dz / 2)), axis=1)
    c_n7 = np.stack(((x - dx / 2), (y + dy / 2), (z + dz / 2)), axis=1)
    c_n8 = np.stack(((x + dx / 2), (y + dy / 2), (z + dz / 2)), axis=1)

    # - Concatenate
    # nodes = np.concatenate((c_n1, c_n2, c_n3, c_n4, c_n5, c_n6, c_n7, c_n8), axis=0)
    nodes = np.hstack((c_n1, c_n2, c_n3, c_n4, c_n5, c_n6, c_n7, c_n8)).ravel().reshape(n_cells * 8, 3)

    # create the cells
    # REF: https://docs/pyvista.org/examples/00-load/create-unstructured-surface.html
    cells_hex = np.arange(n_cells * 8).reshape(n_cells, 8)

    grid = pv.UnstructuredGrid({pv.CellType.VOXEL: cells_hex}, nodes)

    # add the attributes (column) data
    for col in blocks.columns:
        grid.cell_data[col] = blocks[col].values

    return grid


def convert_tensor_to_regular(tensor_model: TensorGridBlockModel) -> RegularBlockModel:
    # Check if the TensorGridBlockModel is regularly spaced
    tensor_geometry = TensorGeometry.from_element(tensor_model)
    if not tensor_geometry.is_regular:
        raise ValueError(
            "The TensorGridBlockModel is not regularly spaced and cannot be converted to a RegularBlockModel.")

    # Create a RegularGeometry from the TensorGeometry
    regular_geometry = RegularGeometry(
        corner=tensor_geometry.corner,
        axis_u=tensor_geometry.axis_u,
        axis_v=tensor_geometry.axis_v,
        axis_w=tensor_geometry.axis_w,
        block_size=(tensor_geometry.tensor_u[0], tensor_geometry.tensor_v[0], tensor_geometry.tensor_w[0]),
        shape=tensor_geometry.shape
    )

    # Create a RegularBlockModel using the RegularGeometry
    regular_model = RegularBlockModel(
        origin=list(regular_geometry.corner),
        axis_u=list(regular_geometry.axis_u),
        axis_v=list(regular_geometry.axis_v),
        axis_w=list(regular_geometry.axis_w),
        block_size=list(regular_geometry.block_size),
        block_count=list(regular_geometry.shape),
        cbc = [1] * regular_geometry.num_cells
    )

    # Copy attributes from the TensorGridBlockModel to the RegularBlockModel
    regular_model.attributes = tensor_model.attributes

    return regular_model


import json
import os
from pathlib import Path
from typing import Optional, Literal, Union

import omf
import pandas as pd
import ydata_profiling

from omfpandas import OMFPandasReader
from omfpandas.audit import ChangeMessage
from omfpandas.base import OMFPandas
from omfpandas.blockmodels.convert_blockmodel import df_to_blockmodel, blockmodel_to_df

from omfpandas.extras import _import_ydata_profiling, _import_pandera, _import_pandera_io
from omfpandas.utils.pandas_utils import parse_vars_from_expr
from omfpandas.utils.pandera_utils import DataFrameMetaProcessor, load_schema_from_yaml
from omfpandas.utils import log_timer

def get_username():
    try:
        return os.getlogin()
    except OSError:
        return os.environ.get('USER') or os.environ.get('USERNAME') or 'unknown_user'

PathLike = Union[str, Path, os.PathLike]

class OMFPandasWriter(OMFPandasReader):
    """A class to write pandas dataframes to an OMF file.

    Methods are named to align with CRUD
    create -> create a new element on the omf file
    read -> read a dataframe from an element (using OMFPandasReader inheritance)
    update -> update (part of) an existing element by providing a dataframe  # TODO
    delete -> delete an element from the omf file.  # TODO

    Attributes:
        filepath (Path): Path to the OMF file.
    """

    def __init__(self, filepath: PathLike):
        """Instantiate the OMFPandasWriter object.

        Args:
            filepath (Path): Path to the OMF file.
        """
        OMFPandas.__init__(self, filepath)
        self.user_id = get_username()

        if not isinstance(filepath, Path):
            filepath = Path(filepath)

        if not filepath.exists():
            # log a message and create a new project
            project = omf.Project()
            project.name = filepath.stem
            project.description = f"OMF file created by OMFPandasWriter: {filepath.name}"
            self._logger.info(f"Creating new OMF file: {filepath}")
            self.project = project  # to enable the write_to_changelog method
            # create the audit record
            self.write_to_changelog(element='None', action='create', description=f"File created: {filepath}")
            # save the (now modified) project to the omf file
            self.persist_project()

        super().__init__(filepath)

    @log_timer()
    def create_blockmodel(self, blocks: pd.DataFrame, blockmodel_name: str,
                          pd_schema: Optional[Union[Path, dict]] = None,
                          allow_overwrite: bool = False):
        """Create an omf BlockModel from a dataframe.

        Only dataframes with centroid (x, y, z) and block dims (dx, dy, dz) indexes are supported.

        Args:
            blocks (pd.DataFrame): The dataframe to write to the BlockModel.
            blockmodel_name (str): The name of the BlockModel to write to. Use dot notation for composite (e.g., Composite.BlockModel).
            pd_schema (Optional[Union[Path, dict]]): The path to the Pandera schema file or a dict of the schema.
             Default is None.  If provided, the schema will be used to validate the dataframe before writing.
            allow_overwrite (bool): If True, overwrite the existing BlockModel. Default is False.

        Raises:
            ValueError: If the element retrieved is not a BlockModel.
        """
        composite_name = None
        full_blockmodel_name = blockmodel_name

        if '.' in blockmodel_name:
            composite_name, blockmodel_name = blockmodel_name.split('.', 1)

        calculation_map: dict = {}
        if pd_schema is not None:
            pa = _import_pandera()
            if not isinstance(pd_schema, (Path, dict)):
                raise ValueError("pd_schema must be a Path to a Pandera schema file or a dict of the schema.")
            elif isinstance(pd_schema, Path) and not pd_schema.exists():
                raise FileNotFoundError(f"Schema file not found: {pd_schema}")
            elif isinstance(pd_schema, dict):
                paio = _import_pandera_io()
                pd_schema = paio.deserialize_schema(pd_schema)
            elif isinstance(pd_schema, Path):
                pd_schema = load_schema_from_yaml(pd_schema)
                # below line suffers from bug: https://github.com/unionai-oss/pandera/issues/1301
                # pd_schema = pa.DataFrameSchema.from_yaml(pd_schema)
                self._logger.info(f"Validating dataframe with schema: {pd_schema}")

            # validate the dataframe, which may modify it via coercion

            # add any calculated attributes in the schema
            dfmp: DataFrameMetaProcessor = DataFrameMetaProcessor(schema=pd_schema)
            calculation_map = dfmp.calculation_map
            blocks = dfmp.preprocess(blocks)
            blocks = dfmp.validate(blocks, return_calculated_columns=False)

            self._logger.info(f"Creating BlockModel from dataframe: {blockmodel_name}")
            bm = df_to_blockmodel(blocks, blockmodel_name)

            # persist the schema inside the omf file
            bm.description = pd_schema.description
            bm.metadata['pd_schema'] = pd_schema.to_json()
        else:
            self._logger.info(f"Creating BlockModel from dataframe: {blockmodel_name}")
            bm = df_to_blockmodel(blocks, blockmodel_name)

        if bm.name in [element.name for element in self.project.elements]:
            if not allow_overwrite:
                raise ValueError(f"BlockModel '{blockmodel_name}' already exists in the OMF file: {self.filepath}.  "
                                 f"If you want to overwrite, set allow_overwrite=True.")
            else:
                # remove the existing volume from the project
                volume_to_remove = [element for element in self.project.elements if element.name == bm.name][0]
                self.project.elements.remove(volume_to_remove)

        log_description: str = f"BlockModel written with {len(bm.attributes)} attributes"
        if composite_name:
            # get the composite if it exists or create it if it does not
            if composite_name not in [element.name for element in self.project.elements]:
                composite = omf.Composite(name=composite_name)
                self.project.elements.append(composite)
            else:
                composite = self.get_element_by_name(composite_name)
            composite.elements.append(bm)
            log_description += f" in composite {composite_name}"
        else:
            self.project.elements.append(bm)

        # create the audit record
        self.write_to_changelog(element=bm.name, action='create', description=log_description)
        # write the omf project to file
        self.persist_project()

        # write the calculated variables to the omf block model metadata
        if pd_schema is not None:
            self.create_calculated_blockmodel_attributes(full_blockmodel_name, calc_definitions=calculation_map)

    def create_calculated_blockmodel_attributes(self, blockmodel_name: str, calc_definitions: dict[str, str]):
        """Create a calculated attribute for a BlockModel.

        Calculated attributes reduce storage space by storing the calculation expression instead of the data.
        When the attribute is accessed, the expression is evaluated and the result returned.
        The calculation expression must be a valid pandas expression, and is stored in the metadata of the
        blockmodel object.  Only OMF2 supports this feature.

        Args:
            blockmodel_name (str): The name of the BlockModel.
            calc_definitions (dict[str, str]): A dictionary of attribute names and calculation expressions.
        """
        bm = self.get_element_by_name(blockmodel_name)

        # confirm the element is a BlockModel
        if bm.__class__.__name__ not in ['RegularBlockModel', 'TensorGridBlockModel']:
            raise ValueError(f"Element '{bm}' is not a supported BlockModel in the OMF file: {self.filepath}")
        for attr_name, expr in calc_definitions.items():
            # check the attribute does not already exist
            if attr_name in self.get_element_attribute_names(blockmodel_name):
                raise ValueError(f"Attribute '{attr_name}' already exists in BlockModel '{blockmodel_name}'.")
            attrs_in_scope = list(set(parse_vars_from_expr(expr)))
            # validate that the attributes in the expression are in the schema
            for attr in attrs_in_scope:
                if attr not in self.get_element_attribute_names(blockmodel_name):
                    raise ValueError(f"Expression attribute '{attr}' not found in BlockModel '{blockmodel_name}'.")

            # Load the head dataset containing the attributes in the expression to validate the expression is valid,
            # though if the attribute exists in the schema file then it will be validated on the entire dataset.
            index_filter = list(range(0, 6))
            if bm.metadata.get('pd_schema'):
                pa = _import_pandera()
                col_schema = pa.io.from_json(bm.metadata['pd_schema']).columns.get(attr_name)
                if col_schema:
                    index_filter = None

            # validate the expression
            df: pd.DataFrame = self.read_blockmodel(blockmodel_name, attributes=attrs_in_scope,
                                                    index_filter=index_filter)
            try:
                df.eval(expr)
            except Exception as e:
                raise ValueError(f"Expression '{expr}' failed during evaluation: {e}")

            # persist the configuration to the blockmodel metadata
            if 'calculated_attributes' in bm.metadata:
                bm.metadata['calculated_attributes'][attr_name] = expr
            else:
                bm.metadata['calculated_attributes'] = {attr_name: expr}

            self.write_to_changelog(element=blockmodel_name, action='create',
                                    description=f"Calculated attribute [{attr_name}] added with expression {expr}")

        self.persist_project()

    def write_to_changelog(self, element: str, action: Literal['create', 'update', 'delete'], description: str):
        """Write a change message to the OMF file.

        Args:
            element: The name of the element that was changed
            action: The action taken on the object
            description: Description of the change

        Returns:

        """

        if 'changelog' not in self.project.metadata:
            self.project.metadata['changelog'] = []
        msg = ChangeMessage(element=element, user=self.user_id, action=action, description=description)
        self.project.metadata['changelog'].append(str(msg))

    def persist_project(self):
        """Persist the omf project to file and reload the project property"""
        omf.save(project=self.project, filename=str(self.filepath), mode='w')
        self.project = omf.load(str(self.filepath))

    def write_blockmodel_attribute(self, blockmodel_name: str, series: pd.Series,
                                   allow_overwrite: bool = False):
        """Write data to a specific attribute of a BlockModel.

        Args:
            blockmodel_name (str): The name of the BlockModel.
            series (pd.Series): The data to write to the attribute.
            allow_overwrite (bool): If True, overwrite the existing attribute. Default is False.
        """
        from omfpandas.blockmodels.attributes import series_to_attribute

        bm = self.get_element_by_name(blockmodel_name)
        if bm.metadata.get('pd_schema'):
            pa = _import_pandera()
            # validate the data
            schema = pa.io.from_json(bm.metadata['pd_schema'])
            series = schema.validate(series.to_frame())
            series = series.iloc[:, 0]  # back to series

        attrs: list[str] = self.get_element_attribute_names(blockmodel_name)
        if series.name in attrs:
            if allow_overwrite:
                # get the index in the list
                attr_pos = attrs.index(str(series.name))
                bm.attributes[attr_pos] = series_to_attribute(series)
            else:
                raise ValueError(f"Attribute '{series.name}' already exists in BlockModel '{blockmodel_name}'.  "
                                 f"If you want to overwrite, set allow_overwrite=True.")
        else:
            bm.attributes.append(series_to_attribute(series))

        self._delete_profile_report(blockmodel_name)

        # todo: re-profile...

        self.write_to_changelog(element=bm.name, action='create', description=f"Attribute [{series.name}] written")

    def delete_blockmodel_attribute(self, blockmodel_name: str, attribute_name: str):
        """Delete an attribute from a BlockModel.

        Args:
            blockmodel_name (str): The name of the BlockModel.
            attribute_name (str): The name of the attribute.
        """
        bm = self.get_element_by_name(blockmodel_name)
        attrs: list[str] = self.get_element_attribute_names(bm)
        if attribute_name in attrs:
            del bm.attributes[attribute_name]
        else:
            raise ValueError(f"Attribute '{attribute_name}' not found in BlockModel '{blockmodel_name}'.")

        self._delete_profile_report(blockmodel_name)

        # create the audit record, which also saves the file
        self.write_to_changelog(element=bm.name, action='delete', description=f"{attribute_name} deleted")

    @log_timer()
    def profile_blockmodel(self, blockmodel_name: str, query: Optional[str] = None):
        """Profile a BlockModel.

        Profiling will be skipped if the data has not changed.

        Args:
            blockmodel_name (str): The name of the BlockModel to profile.
            query (Optional[str]): A query to filter the data before profiling.

        Returns:
            pd.DataFrame: The profiled data.
        """

        _import_ydata_profiling()

        df: pd.DataFrame = OMFPandasReader(self.filepath).read_blockmodel(blockmodel_name, query=query)
        el = self.get_element_by_name(blockmodel_name)
        bm_type = str(type(el)).split('.')[-1].rstrip("'>")
        dataset: dict = {"description": f"{el.description} Filter: {query if query else 'no_filter'}",
                         "creator": self.user_id, "url": self.filepath.as_uri()}
        column_descriptions: dict = {}
        if el.metadata.get('pd_schema'):
            column_defs: dict = json.loads(el.metadata['pd_schema'])['columns']
            column_descriptions = {k: f"{v['title']}: {v['description']}" for k, v in column_defs.items()}

        profile: ydata_profiling.ProfileReport = df.profile_report(title=f"{el.name} {bm_type}", dataset=dataset,
                                                                   variables={"descriptions": column_descriptions})

        # persist the profile report as html to the omf file, larger but cannot serialise the ProfileReport object,
        # nor recreate the report from json.
        d_profile: dict = {query if query else 'no_filter': profile.to_html()}

        if el.metadata.get('profile'):
            el.metadata['profile'] = {**el.metadata['profile'], **d_profile}
        else:
            el.metadata['profile'] = d_profile

        self.write_to_changelog(element=blockmodel_name, action='create',
                                description=f"Profiled with query {query}")

        return profile

    def write_block_model_schema(self, blockmodel_name: str, pd_schema_filepath: Path):
        """Write a Pandera schema to the OMF file.

        Args:
            blockmodel_name (str): The name of the BlockModel.
            pd_schema_filepath (Path): The path to the Pandera schema yaml file.
        """
        pa = _import_pandera()
        bm = self.get_element_by_name(blockmodel_name)
        pd_schema = pa.DataFrameSchema.from_yaml(pd_schema_filepath)
        bm.metadata['pd_schema'] = pd_schema.to_json()

        el = self.get_element_by_name(blockmodel_name)
        schema_title = pd_schema.title if pd_schema.title else ''
        schema_description = pd_schema.description if pd_schema.description else ''
        el.description = f"{schema_title}: {schema_description}"

        self.write_to_changelog(element=blockmodel_name, action='create', description=f"Schema written")

    def _delete_profile_report(self, blockmodel_name: str):
        """Delete the profile report from the OMF file when data has changed."""
        bm = self.get_element_by_name(blockmodel_name)

        if 'profile' in bm.metadata:
            del bm.metadata['profile']

    def blockmodel_to_parquet(self, blockmodel_name: str, out_path: Optional[Path] = None,
                              variables: Optional[list[str]] = None,
                              allow_overwrite: bool = False):
        """Convert blockmodel to a Parquet file.

        Args:
            blockmodel_name (str): The BlockModel element to convert.
            out_path (Optional[Path]): The path to the Parquet file to write. If None, a file with the blockmodel name is
            created.
            variables (Optional[list[str]]): The variables to include in the DataFrame. If None, all variables are included.
            allow_overwrite (bool): If True, overwrite the existing Parquet file. Default is False.

        Raises:
            FileExistsError: If the file already exists and allow_overwrite is False.
        """
        bm = self.get_element_by_name(blockmodel_name)
        if out_path is None:
            out_path = Path(f"{blockmodel_name}.parquet")
        if out_path.exists() and not allow_overwrite:
            raise FileExistsError(
                f"File already exists: {out_path}. If you want to overwrite, set allow_overwrite=True.")
        df: pd.DataFrame = blockmodel_to_df(blockmodel=bm, variables=variables)
        df.to_parquet(out_path)

    def blockmodel_to_orc(self, blockmodel_name: str, out_path: Optional[Path] = None,
                          variables: Optional[list[str]] = None,
                          allow_overwrite: bool = False):
        """Convert blockmodel to an ORC file.

        Args:
            blockmodel_name (str): The BlockModel element to convert.
            out_path (Optional[Path]): The path to the ORC file to write. If None, a file with the blockmodel name is
            created.
            variables (Optional[list[str]]): The variables to include in the DataFrame. If None, all variables are included.
            allow_overwrite (bool): If True, overwrite the existing ORC file. Default is False.

        Raises:
            FileExistsError: If the file already exists and allow_overwrite is False.
        """
        bm = self.get_element_by_name(blockmodel_name)

        if out_path is None:
            out_path = Path(f"{blockmodel_name}.orc")
        if out_path.exists() and not allow_overwrite:
            raise FileExistsError(
                f"File already exists: {out_path}. If you want to overwrite, set allow_overwrite=True.")
        is_tensor: bool = True if bm.__class__.__name__ == 'TensorGridBlockModel' else False
        df: pd.DataFrame = blockmodel_to_df(blockmodel=bm, variables=variables)
        df.to_orc(out_path)



from pathlib import Path

import yaml
from pandera import DataFrameSchema, Column, Check
from pandera.engines import pandas_engine
import pandas as pd
from pandera.io import _deserialize_check_stats


class DataFrameMetaProcessor:
    def __init__(self, schema: DataFrameSchema):
        self.schema: DataFrameSchema = schema
        self.supported_column_meta_keys = ['alias', 'calculation', 'decimals']

    @property
    def alias_map(self):
        return {col.metadata['alias']: col_name for col_name, col in self.schema.columns.items() if
                col.metadata and 'alias' in col.metadata}

    @property
    def calculation_map(self):
        return {col_name: col.metadata['calculation'] for col_name, col in self.schema.columns.items() if
                col.metadata and 'calculation' in col.metadata}

    @property
    def decimals_map(self):
        return {col_name: col.metadata['decimals'] for col_name, col in self.schema.columns.items() if
                col.metadata and 'decimals' in col.metadata}

    def rename_from_meta_alias(self, df: pd.DataFrame) -> pd.DataFrame:
        return df.rename(columns=self.alias_map)

    def calculate_from_meta_calculation(self, df: pd.DataFrame) -> pd.DataFrame:
        for col_name, calculation in self.calculation_map.items():
            df[col_name] = eval(calculation, {}, df.to_dict('series'))
        return df

    def round_to_decimals(self, df: pd.DataFrame, columns: list = None) -> pd.DataFrame:
        if columns is None:
            columns = self.decimals_map.keys()
        for col_name in columns:
            if col_name in self.decimals_map:
                df[col_name] = df[col_name].round(self.decimals_map[col_name])
        return df

    def preprocess(self, df: pd.DataFrame) -> pd.DataFrame:
        df = self.rename_from_meta_alias(df)
        df = self.round_to_decimals(df)
        df = self.calculate_from_meta_calculation(df)
        df = self.round_to_decimals(df, columns=list(self.calculation_map.keys()))
        return df

    def validate(self, df: pd.DataFrame, return_calculated_columns: bool = True) -> pd.DataFrame:
        df = self.schema.validate(df)
        if not return_calculated_columns:
            return df.drop(columns=list(self.calculation_map.keys()))
        return df


def load_schema_from_yaml(yaml_path: Path) -> DataFrameSchema:
    """Load a DataFrameSchema from a YAML file."""
    with open(yaml_path, "r", encoding="utf-8") as f:
        schema_dict = yaml.safe_load(f)

    columns = {
        col_name: Column(**_deserialize_component_stats(col_stats))
        for col_name, col_stats in schema_dict["columns"].items()
    }

    return DataFrameSchema(
        columns=columns,
        checks=schema_dict.get("checks"),
        index=schema_dict.get("index"),
        dtype=schema_dict.get("dtype"),
        coerce=schema_dict.get("coerce", False),
        strict=schema_dict.get("strict", False),
        name=schema_dict.get("name", None),
        ordered=schema_dict.get("ordered", False),
        unique=schema_dict.get("unique", None),
        report_duplicates=schema_dict.get("report_duplicates", "all"),
        unique_column_names=schema_dict.get("unique_column_names", False),
        add_missing_columns=schema_dict.get("add_missing_columns", False),
        title=schema_dict.get("title", None),
        description=schema_dict.get("description", None),
    )


def _deserialize_component_stats(serialized_component_stats):
    dtype = serialized_component_stats.get("dtype")
    if dtype:
        dtype = pandas_engine.Engine.dtype(dtype)

    description = serialized_component_stats.get("description")
    title = serialized_component_stats.get("title")

    checks = serialized_component_stats.get("checks")
    if checks is not None:
        checks = [
            _deserialize_check_stats(
                getattr(Check, check_name), check_stats, dtype
            )
            for check_name, check_stats in checks.items()
        ]

    return {
        "title": title,
        "description": description,
        "dtype": dtype,
        "checks": checks,
        **{
            key: serialized_component_stats.get(key)
            for key in [
                "name",
                "nullable",
                "unique",
                "coerce",
                "required",
                "regex",
                "metadata"
            ]
            if key in serialized_component_stats
        },
    }


"""
Functions to support coordinate encoding/decoding for block models.
NOTE: The encoding scheme can only support x, y, z, given the 64 bit limit, so only regular block models are supported.
"""
import pandas as pd


def encode_coordinates(x: float, y: float, z: float) -> int:
    """Encode the coordinates into a 64-bit integer."""
    x_int = int(x * 10) & 0xFFFFFF
    y_int = int(y * 10) & 0xFFFFFF
    z_int = int(z * 10) & 0xFFFF
    encoded = (x_int << 40) | (y_int << 16) | z_int
    return encoded


def decode_coordinates(encoded: int) -> tuple:
    """Decode the 64-bit integer back to the original coordinates."""
    x_int = (encoded >> 40) & 0xFFFFFF
    y_int = (encoded >> 16) & 0xFFFFFF
    z_int = encoded & 0xFFFF
    x = x_int / 10.0
    y = y_int / 10.0
    z = z_int / 10.0
    return x, y, z


def multiindex_to_encoded_index(multi_index: pd.MultiIndex) -> pd.Index:
    """Convert a MultiIndex to an encoded integer Index."""
    encoded_indices = [
        encode_coordinates(x, y, z)
        for x, y, z in zip(
            multi_index.get_level_values("x"),
            multi_index.get_level_values("y"),
            multi_index.get_level_values("z"),
        )
    ]
    return pd.Index(encoded_indices, name='encoded_xyz')


def encoded_index_to_multiindex(encoded_index: pd.Index) -> pd.MultiIndex:
    """Convert an encoded integer Index back to a MultiIndex."""
    decoded_coords = [decode_coordinates(encoded) for encoded in encoded_index]
    x, y, z = zip(*decoded_coords)
    return pd.MultiIndex.from_arrays([x, y, z], names=["x", "y", "z"])


import os
from pathlib import Path
from typing import Optional, Union

import pandas as pd

from omfpandas.base import OMFPandas, SUPPORTED_BM_TYPES
from omfpandas.blockmodel import OMFBlockModel
from omfpandas.blockmodels import multiindex_to_encoded_index
from omfpandas.blockmodels.convert_blockmodel import blockmodel_to_df
from omfpandas.blockmodels.geometry import Geometry
from omfpandas.utils.pandas_utils import parse_vars_from_expr

PathLike = Union[str, Path, os.PathLike]


class OMFPandasReader(OMFPandas):
    """A class to read an OMF file to a pandas DataFrame.

    Attributes:
        filepath (Path): Path to the OMF file.

    """

    def __init__(self, filepath: PathLike):
        """Instantiate the OMFPandasReader object

        Args:
            filepath: Path to the OMF file.
        """
        if not isinstance(filepath, Path):
            filepath = Path(filepath)

        if not filepath.exists():
            raise FileNotFoundError(f"File does not exist: {filepath}")
        super().__init__(filepath)

    def read_blockmodel(
            self,
            blockmodel_name: str,
            attributes: Optional[list[str]] = None,
            query: Optional[str] = None,
            index_filter: Optional[list[int]] = None,
            encode_index: bool = False,
    ) -> pd.DataFrame:
        """Return a DataFrame from a BlockModel.

        Only variables assigned to the `cell` (as distinct from the grid `points`) are loaded.

        Args:
            blockmodel_name (str): The name of the BlockModel to read. Use dot notation for composite (e.g., Composite.BlockModel).
            attributes (Optional[list[str]]): The attributes/variables to include in the DataFrame. If None, all
                variables are included.
            query (Optional[str]): A query string to filter the DataFrame. Default is None.
            index_filter (Optional[list[int]]): A list of indexes to filter the DataFrame. Default is None.
            encode_index (bool): If True, encode the index to a single integer.

        Returns:
            pd.DataFrame: The DataFrame representing the BlockModel.
        """
        bm = self.get_element_by_name(blockmodel_name)
        # check the element retrieved is the expected type
        if bm.__class__.__name__ not in ["RegularBlockModel", "TensorGridBlockModel"]:
            raise ValueError(
                f"Element '{bm}' is not a supported BlockModel in the OMF file: {self.filepath}"
            )
        res: pd.DataFrame = blockmodel_to_df(
            bm, variables=attributes, query=query, index_filter=index_filter
        )
        if encode_index:
            res.index = multiindex_to_encoded_index(res.index)
        return res

    def read_block_models(
            self,
            blockmodel_attributes: dict[str, list[str]],
            query: Optional[str] = None,
            encode_index: bool = False,
    ) -> pd.DataFrame:
        """Return a DataFrame from multiple BlockModels.

        Args:
            blockmodel_attributes (dict[str, list[str]]): A dictionary of BlockModel names and the variables to include.
                If the dict value is None, all attributes in the blockmodel (key) are included.
            query (Optional[str]): A query string to filter the DataFrame. Default is None.
            encode_index: If True, encode the index to a single integer.

        Returns:
            pd.DataFrame: The DataFrame representing the merged BlockModels.
        """
        block_models: dict[str, pd.DataFrame] = {}
        geometry_indexes: dict[str, pd.MultiIndex] = {}
        index_filter = None
        if query:
            # we use the query to generate an index_filter to filter the block models since they may not have
            # the query attributes available in the given model.
            query_attrs: list[str] = parse_vars_from_expr(query)
            # extract the attributes for the query_attributes, noting that they may be in different block models
            chunks: list[pd.Series] = []
            for bm_name in blockmodel_attributes:
                for attr in query_attrs:
                    if attr in self.blockmodel_attributes[bm_name]:
                        chunks.append(
                            self.read_blockmodel(blockmodel_name=bm_name,
                                                 attributes=[attr], encode_index=encode_index)[attr])
            tmp_df: pd.DataFrame = pd.concat(chunks, axis=1).query(query)
            # get the index locations of the filtered index relative to the full index
            filtered_index = tmp_df.index
            # full_index = geometry_to_index(self.get_element_by_name(self._elements[0].name).geometry)
            full_index: pd.MultiIndex = self.get_bm_geometry(
                blockmodel_name=list(blockmodel_attributes.keys())[0]
            ).to_multi_index()
            # Find the intersection of the two MultiIndex objects
            intersection = full_index.intersection(filtered_index)
            index_filter = full_index.get_indexer(intersection)

        for bm_name, requested_attrs in blockmodel_attributes.items():
            # check that the requested attrs exist in the specified bm
            available_attrs = self.blockmodel_attributes[bm_name]
            if requested_attrs is None:
                requested_attrs = available_attrs
            else:
                missing_attrs = set(requested_attrs) - set(available_attrs)
                if missing_attrs:
                    raise ValueError(
                        f"Attributes {missing_attrs} not found in BlockModel '{bm_name}'. "
                        f"Available attributes are: {available_attrs}"
                    )

            block_models[bm_name] = self.read_blockmodel(
                blockmodel_name=bm_name,
                attributes=requested_attrs,
                index_filter=index_filter if query else None,
            )
            # geometry_indexes[bm_name] = geometry_to_index(self.get_bm_geometry(bm_name))
            geometry_indexes[bm_name] = self.get_bm_geometry(
                blockmodel_name=bm_name
            ).to_multi_index()

        # validate the indexes are equivalent
        def ensure_identical_indexes(index_dict: dict[str, pd.MultiIndex]) -> None:
            if not index_dict:
                return

            first_index = next(iter(index_dict.values()))
            for name, index in index_dict.items():
                if not first_index.equals(index):
                    raise ValueError(
                        f"Index for '{name}' is different from the first index."
                    )

        ensure_identical_indexes(geometry_indexes)

        return pd.concat(block_models.values(), axis=1)

    def plot_blockmodel(
            self,
            blockmodel_name: str,
            scalar: str,
            threshold: bool = True,
            show_edges: bool = True,
            show_axes: bool = True,
    ) -> "pv.Plotter":
        """Plot the BlockModel using PyVista.

        Args:
            blockmodel_name (str): The name of the BlockModel to plot.
            scalar (str): The scalar to plot.
            threshold (bool): If True, plot the thresholded mesh. Default is True.
            show_edges (bool): If True, show the edges. Default is True.
            show_axes (bool): If True, show the axes. Default is True.

        Returns:
            pv.Plotter: The PyVista plotter object.
        """
        block_model = OMFBlockModel(self.get_element_by_name(blockmodel_name))
        return block_model.plot(
            scalar=scalar,
            threshold=threshold,
            show_edges=show_edges,
            show_axes=show_axes,
        )

    def find_nearest_centroid(
            self, x: float, y: float, z: float, blockmodel_name: Optional[str] = None
    ) -> tuple[float, float, float]:
        """Find the nearest centroid for provided x, y, z points using a math rounding approach considering the reference centroid.

        Args:
            x (float): X coordinate.
            y (float): Y coordinate.
            z (float): Z coordinate.
            blockmodel_name: The optional block model name.  If not provided, the geometry for the first Tensor
             block model is used.

        Returns:
            Tuple[float, float, float]: The coordinates of the nearest centroid.
        """

        # get the geometry for the first block model if not provided
        if blockmodel_name is None:
            blockmodel_names = [
                element_name
                for element_name, element_type in self.element_types.items()
                if element_type in SUPPORTED_BM_TYPES
            ]
            if not blockmodel_names:
                raise ValueError("No BlockModel found in the OMF file.")
            blockmodel_name = blockmodel_names[0]

        # get the geometry for the block model
        geometry: Geometry = OMFBlockModel(
            self.get_element_by_name(blockmodel_name)
        ).geometry
        # perform the lookup
        nearest_x, nearest_y, nearest_z = geometry.nearest_centroid_lookup(x, y, z)

        return nearest_x, nearest_y, nearest_z



import logging
from typing import Union, Optional

import pandas as pd

from omfpandas.blockmodels.convert_blockmodel import blockmodel_to_df, df_to_blockmodel, df_to_pv_structured_grid, \
    df_to_pv_unstructured_grid
from omfpandas.blockmodels.geometry import Geometry, TensorGeometry, RegularGeometry


class OMFBlockModel:
    def __init__(self, blockmodel: Union['BaseBlockModel', 'RegularBlockModel', 'TensorGridBlockModel']):
        self._logger = logging.getLogger(__class__.__name__)
        self.blockmodel = blockmodel
        self.bm_type: str = blockmodel.__class__.__name__
        self.geometry: Geometry = TensorGeometry.from_element(
            blockmodel) if self.bm_type == 'TensorGridBlockModel' else RegularGeometry.from_element(blockmodel)
        self.attributes = {a.name: a.schema.split('.')[-1] for a in self.blockmodel.attributes}

    def to_dataframe(self, variables: Optional[list[str]] = None, query: Optional[str] = None,
                     index_filter: Optional[list[int]] = None) -> pd.DataFrame:
        return blockmodel_to_df(blockmodel=self.blockmodel,
                                variables=variables,
                                query=query,
                                index_filter=index_filter)

    @classmethod
    def from_dataframe(cls, df: pd.DataFrame, blockmodel_name: str):
        return cls(blockmodel=df_to_blockmodel(df=df, blockmodel_name=blockmodel_name))

    def plot(self, scalar: str, threshold: bool=True, show_edges: bool = True, show_axes: bool=True) -> 'pv.Plotter':
        import pyvista as pv
        if scalar not in self.attributes.keys():
            raise ValueError(f"Column '{scalar}' not found in the OMFBlockModel.")

        # Create a PyVista plotter
        plotter = pv.Plotter()

        mesh = self.get_blocks(attributes=[scalar])

        # Add a thresholded mesh to the plotter
        if threshold:
            plotter.add_mesh_threshold(mesh, scalars=scalar, show_edges=show_edges)
        else:
            plotter.add_mesh(mesh, scalars=scalar, show_edges=show_edges)

        plotter.title = self.blockmodel.name
        if show_axes:
            plotter.show_axes()

        return plotter

    def get_blocks(self, attributes: Optional[list[str]] = None) -> Union['pv.StructuredGrid', 'pv.UnstructuredGrid']:

        if attributes is None:
            attributes = list(self.attributes.keys())
        df = self.to_dataframe(variables=attributes)

        try:
            # Attempt to create a regular grid
            grid = df_to_pv_structured_grid(df)
            self._logger.debug("Created a pv.StructuredGrid.")
        except ValueError:
            # If it fails, create an irregular grid
            grid = df_to_pv_unstructured_grid(df)
            self._logger.debug("Created a pv.UnstructuredGrid.")
        return grid


